# -*- coding: utf-8 -*-

"""Settings module."""

import importlib

from django.conf import settings


def _import(complete_path):
    module_name = '.'.join(complete_path.split('.')[:-1])
    module = importlib.import_module(name=module_name)
    function_or_class = getattr(module, complete_path.split('.')[-1])
    return function_or_class


def _type_checker(t):
    def check(name, value):
        """Check value for given type."""
        if not isinstance(value, t):
            raise ValueError('%s must be %s, not %s' % (
                name, t, value.__class__))
    check.__name__ = 'check_%s' % t.__name__
    return check


def _type_tuple_checker(ty, tu):
    # pylama:ignore=C0111
    def check(name, value):
        if not isinstance(value, tu):
            raise ValueError('%s must be %s, not %s' % (
                name, tu, value.__class__))
        if not all(isinstance(o, ty) for o in value):
            raise ValueError('All elements of %s must be %s' % (name, ty))
    check.__name__ = 'check_%s_%s' % (ty.__name__, tu.__name__)
    return check


check_string = _type_checker(str)
check_int = _type_checker(int)
check_bool = _type_checker(bool)
check_float = _type_checker(float)
check_list = _type_checker(list)
check_dict = _type_checker(dict)
check_set = _type_checker(set)


def check_positive_int(name, value):
    """Check positive int value."""
    if not isinstance(value, int):
        raise ValueError('%s must be int' % name)
    elif value < 0:
        raise ValueError('%s must be positive or zero' % name)


def check_positive_float(name, value):
    """Check positive int value."""
    if not isinstance(value, float):
        raise ValueError('%s must be float' % name)
    elif value < 0.0:
        raise ValueError('%s must be positive or zero' % name)


check_string_list = _type_tuple_checker(str, list)
check_string_set = _type_tuple_checker(str, set)
check_int_list = _type_tuple_checker(str, list)
check_int_set = _type_tuple_checker(str, set)
check_bool_list = _type_tuple_checker(str, list)
check_bool_set = _type_tuple_checker(str, set)
check_float_list = _type_tuple_checker(str, list)
check_float_set = _type_tuple_checker(str, set)


class Setting(object):
    """
    Generic setting class.

    Serve as base class for more specific setting types.
    """

    def __init__(self,
                 name=None,
                 default=None,
                 checker=lambda n, v: None,
                 transformer=lambda v: v,
                 prefix=None):
        """
        Initialization method.

        Args:
            name (str): name of the setting.
            default (obj): default value given to the setting.
            checker (func):
                function to check the setting. It must take 2 arguments: name,
                value, and raise an error if value is incorrect. Default:
                do nothing.
            transformer ():
                function to transform the value retrieved from Django settings.
                It must take 1 argument: value, and return it transformed.
                Default: identity.
            prefix (str):
                prefix of the setting.
                Will override ``AppSettings.Meta`` prefix.
        """
        self.name = name
        self.default = default
        self.transformer = transformer
        self.checker = checker
        self.prefix = prefix

    def get_raw(self):
        """Get the setting from ``django.conf.settings``."""
        setting_name = self.prefix.upper() + self.name.upper()
        return getattr(settings, setting_name, self.default)

    def get(self):
        """Get the setting and return it transformed."""
        return self.transform()

    def check(self):
        """Check the setting. Raise exception if incorrect."""
        return self.checker(self.name, self.get_raw())

    def transform(self):
        """Get the setting and return it transformed."""
        return self.transformer(self.get_raw())


def _type_setting(cls_name, check_func):
    class _Setting(Setting):
        def __init__(self,
                     name=None,
                     default=None,
                     checker=check_func,
                     transformer=lambda v: v,
                     prefix=None):
            super(_Setting, self).__init__(
                name, default, checker, transformer, prefix)
    _Setting.__name__ = '%sSetting' % cls_name
    return _Setting


StringSetting = _type_setting('String', check_string)
IntSetting = _type_setting('Int', check_int)
PositiveIntSetting = _type_setting('PositiveInt', check_positive_int)
BoolSetting = _type_setting('Bool', check_bool)
FloatSetting = _type_setting('Float', check_float)
PositiveFloatSetting = _type_setting('PositiveFloat', check_positive_float)
ListSetting = _type_setting('List', check_list)
SetSetting = _type_setting('Set', check_set)
DictSetting = _type_setting('Dict', check_dict)

StringListSetting = _type_setting('StringList', check_string_list)
StringSetSetting = _type_setting('StringSet', check_string_set)
IntListSetting = _type_setting('IntList', check_int_list)
IntSetSetting = _type_setting('IntSet', check_int_set)
BoolListSetting = _type_setting('BoolList', check_bool_list)
BoolSetSetting = _type_setting('BoolSet', check_bool_set)
FloatListSetting = _type_setting('FloatList', check_float_list)
FloatSetSetting = _type_setting('FloatSet', check_float_set)


class ImportedObjectSetting(Setting):
    """Setting to import an object given its Python path (a.b.c)."""

    def __init__(self,
                 name=None,
                 default=None,
                 checker=check_string,
                 transformer=_import,
                 prefix=None):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (obj): the default value given to the setting.
            checker (func): the function to check the setting value.
            transformer (func): the function to transform the setting value.
            prefix (str): the setting's prefix.
        """
        super(ImportedObjectSetting, self).__init__(
            name, default, checker, transformer, prefix)



# -*- coding: utf-8 -*-

"""Main test script."""
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings

import pytest

import appsettings


def check_int(name, value):
    """Example checker function."""
    if not isinstance(value, int):
        raise ValueError('%s must be int' % name)


def transform(value):
    """Example transformer function."""
    if value:
        return 'OK'
    return 'NOT OK'


class AppSettingsExample(appsettings.AppSettings):
    """Example AppSettings class."""

    s1 = appsettings.Setting()
    s2 = appsettings.Setting(name='hey')
    s3 = appsettings.Setting(default=1)
    s4 = appsettings.Setting(checker=check_int)
    s5 = appsettings.Setting(transformer=transform)
    s6 = appsettings.Setting(prefix='other')

    class Meta:
        """AppSettings' Meta class."""

        setting_prefix = 'prefix_'


class MainTestCase(TestCase):
    """Main Django test case."""

    def setUp(self):
        """Setup method."""
        self.package = appsettings

    def test_hasttr(self):
        """Main test method."""
        settings = AppSettingsExample()
        for setting in ('s1', 's2', 's3', 's4', 's5'):
            assert hasattr(AppSettingsExample, 'get_%s' % setting)
            assert callable(getattr(AppSettingsExample, 'get_%s' % setting))
            assert hasattr(AppSettingsExample, 'check_%s' % setting)
            assert callable(getattr(AppSettingsExample, 'check_%s' % setting))
            assert hasattr(settings, setting)

    def test_get_methods(self):
        """Test static get methods."""
        assert AppSettingsExample.get_s1() is None
        assert AppSettingsExample.get_s2() is None
        assert AppSettingsExample.get_s3() == 1
        assert AppSettingsExample.get_s4() is None
        assert AppSettingsExample.get_s5() == 'NOT OK'
        assert AppSettingsExample.get_s6() is None

    def test_check_methods(self):
        """Test static check methods."""
        assert AppSettingsExample.check_s1() is None
        with pytest.raises(ValueError):
            AppSettingsExample.check_s4()
        with pytest.raises(ImproperlyConfigured):
            AppSettingsExample.check()
        assert appsettings.AppSettings.check() is None

    def test_get_return_equals_instance_attr(self):
        """Test static get methods return same thing as instance attributes."""
        settings = AppSettingsExample()
        assert AppSettingsExample.get_s1() == settings.s1
        assert AppSettingsExample.get_s2() == settings.s2
        assert AppSettingsExample.get_s3() == settings.s3
        assert AppSettingsExample.get_s4() == settings.s4
        assert AppSettingsExample.get_s5() == settings.s5
        assert AppSettingsExample.get_s6() == settings.s6

    def test_independent_from_order(self):
        """Test get methods do not depend on instantiation/import order."""
        got = [AppSettingsExample.get_s1(),
               AppSettingsExample.get_s2(),
               AppSettingsExample.get_s3(),
               AppSettingsExample.get_s4(),
               AppSettingsExample.get_s5(),
               AppSettingsExample.get_s6()]
        settings = AppSettingsExample()
        assert got == [settings.s1,
                       settings.s2,
                       settings.s3,
                       settings.s4,
                       settings.s5,
                       settings.s6]

    @override_settings(PREFIX_S5=True)
    def test_transform(self):
        """Test transform function is actually called."""
        assert AppSettingsExample.get_s5() == 'OK'

    def test_setting_methods(self):
        """Test setting objects methods."""
        assert isinstance(AppSettingsExample.s1, appsettings.Setting)
        assert AppSettingsExample.s1.get() == AppSettingsExample.get_s1()

    def test_setting_type(self):
        """Test that settings are correctly initialized."""
        from appsettings.settings import (
            IntSetting, DictSetting, check_int, check_dict)
        int_setting = IntSetting()
        dict_setting = DictSetting()
        assert int_setting.checker == check_int
        assert dict_setting.checker == check_dict

    def tearDown(self):
        """Tear down method."""
        del self.package



# -*- coding: utf-8 -*-

u"""Django AppSettings package."""

from django.core.exceptions import ImproperlyConfigured

import six

from .settings import Setting


__version__ = '0.2.1'


class _Metaclass(type):
    """``AppSettings``'s metaclass."""

    def __new__(mcs, cls, bases, dct):
        """
        New method.

        Args:
            cls (str): class name.
            bases (tuple): base classes to inherit from.
            dct (dict): class attributes.

        Returns:
            class: the new created class.
        """
        new_attr = {}
        _meta = dct.pop('Meta', type('Meta', (), {'setting_prefix': ''}))()
        _meta.settings = []

        for name, val in dct.items():
            if isinstance(val, Setting):
                _meta.settings.append(name)
                # populate name
                if val.name is None:
                    val.name = name.upper()
                # populate prefix
                if val.prefix is None:
                    val.prefix = _meta.setting_prefix
                # add getter
                new_attr['get_%s' % name] = val.get
                # add checker
                new_attr['check_%s' % name] = val.check
            new_attr[name] = val
        new_attr['_meta'] = _meta

        return super(_Metaclass, mcs).__new__(mcs, cls, bases, new_attr)

    def __init__(cls, name, bases, dct):
        """
        Initialization method.

        Args:
            name (str): class name.
            bases (tuple): base classes to inherit from.
            dct (dict): class attributes.
        """
        super(_Metaclass, cls).__init__(name, bases, dct)


class AppSettings(six.with_metaclass(_Metaclass)):
    """Base class for application settings."""

    def __init__(self):
        """
        Initialization method.

        Instantiation will replace every class-declared settings with
        the result of their ``get`` method.
        """
        for setting in self._meta.settings:
            setattr(self, setting, getattr(
                self.__class__, 'get_%s' % setting)())

    @classmethod
    def check(cls):
        """
        Class method to check every settings.

        Will raise an ``ImproperlyConfigured`` exception with explanation.
        """
        exceptions = []
        for setting in cls._meta.settings:
            try:
                getattr(cls, 'check_%s' % setting)()
            # pylama:ignore=W0703
            except Exception as e:
                exceptions.append(str(e))
        if exceptions:
            raise ImproperlyConfigured('\n'.join(exceptions))
